@@ -24,17 +24,23 @@
 //!
 //! at your option.
 
+#![forbid(unsafe_code)]
+
 mod cli;
 mod config_file;
 
+#[macro_use]
+extern crate log;
+
 use clap::{ArgEnum, Parser};
 use human_panic::setup_panic;
-use std::io::ErrorKind;
+use indexmap::IndexMap;
+use std::io::{ErrorKind, Read};
 use std::path::Path;
 use std::{fs, path::PathBuf, process::exit, string::String};
 
 use crate::cli::{Cli, Shell};
-use crate::config_file::{ConfigFile, EnvValue, EnvironmentVariables};
+use crate::config_file::{ConfigFile, EnvVariableOption, EnvVariableValue, EnvironmentVariables};
 
 fn main() {
     //! Main function.
@@ -46,11 +52,33 @@ fn main() {
     // Parse the commandline options.
     let cli_options: Cli = Cli::parse();
 
-    // If --toml was specified, use that. Otherwise, get the file and read from it.
-    let (toml_string, file_name) = match cli_options.toml {
-        None => read_config_file(&cli_options),
-        Some(toml) => (toml, String::from("<INPUT>")),
+    // Setup logging
+    env_logger::Builder::new()
+        .filter_level(cli_options.verbose.log_level_filter())
+        .format_timestamp(None)
+        .format_module_path(false)
+        .format_target(false)
+        .format_indent(Some(8)) // Aligns the first line with the other lines
+        .init();
+
+    // Pipe if `cli_options.pipe` is used or if `cli_options.file` is used and equal to "-".
+    let pipe = cli_options.pipe
+        || cli_options
+            .file
+            .as_ref()
+            .map_or(false, |x| x.to_string_lossy() == "-");
+
+    let (toml_string, file_name) = if pipe {
+        // If --pipe was specified, use that as the direct toml.
+        (read_stdin(), String::from("<STDIN>"))
+    } else if let Some(toml) = cli_options.toml {
+        // If --toml was specified, use that. Otherwise, get the file and read from it.
+        (toml, String::from("<STRING>"))
+    } else {
+        // Otherwise, read from the chosen file.
+        read_config_file(&cli_options)
     };
+    info!("Reading file data from {}", file_name);
 
     // Load file data from the TOML file.
     let file_data = match ConfigFile::load(toml_string) {
@@ -59,12 +87,12 @@ fn main() {
         // The file isn't a valid TOML format!
         Err(e) => {
             // Display the error and exit.
-            eprintln!(
-                 "The file {} is not in a valid TOML format or is not in the form Xshe is expecting.",
+            error!(
+                 "The file {} is not in a valid TOML format,\nor it is not in the form Xshe is expecting.",
                  file_name,
              );
             if let Some((line, column)) = e.line_col() {
-                eprintln!("Parse error at line {:}, column {:}", line + 1, column + 1);
+                error!("Parse error at line {:}, column {:}", line + 1, column + 1);
             }
             exit(exitcode::CONFIG)
         }
@@ -72,13 +100,36 @@ fn main() {
 
     let shell: Shell = cli_options.shell;
 
-    // Output the file data converted to the correct shell format to the standard output.
-    let general_output = to_shell_source(&file_data.vars, &shell);
-    print!("{}", general_output);
+    // Deprecation warning
+    if file_data.shell.is_some() {
+        warn!(
+            "Using [shell.SHELL] notation is deprecated\n\
+            See https://github.com/superatomic/xshe/issues/30\n\
+            To be removed in release v1.0.0"
+        );
+    }
 
+    // Output the file data converted to the correct shell format to the standard output.
+    let output = to_shell_source(&file_data.vars, &shell);
+    print!("{}", output);
+
+    // Retain compatibility with deprecated https://github.com/superatomic/xshe/issues/30
+    deprecated_to_specific_shell_source(&file_data, &shell);
+}
+
+fn deprecated_to_specific_shell_source(file_data: &ConfigFile, shell: &Shell) {
     // Output the any specific variables for the shell the same way, if they exist.
-    if let Some(specific_vars) = get_specific_shell(&shell, &file_data) {
-        let shell_specific_output = to_shell_source(specific_vars, &shell);
+    // This behavior is deprecated.
+    if let Some(specific_vars) = get_specific_shell(shell, file_data) {
+        // wrap `specific_vars` to be compatible with the changed `to_shell_source` function.
+        // This is a hack, but it's only to preserve deprecated behavior until it is removed.
+        let wrap_specific_vars = specific_vars
+            .into_iter()
+            .map(|(key, value)| (key.to_owned(), EnvVariableOption::General(value.to_owned())))
+            .collect();
+
+        let shell_specific_output = to_shell_source(&wrap_specific_vars, shell);
+
         print!("{}", shell_specific_output);
     };
 }
@@ -110,22 +161,36 @@ fn display_file_error(kind: ErrorKind, cli_options: &Cli, file: &Path) -> i32 {
                 Some(_) => "Is `--file` set correctly?",
             };
 
-            eprintln!(
-                "The file {:?} does not exist or is a directory. {}",
-                file, help_msg,
-            );
+            error!("The file {:?} does not exist\n{}", file, help_msg);
             exitcode::NOINPUT
         }
 
+        // Unstable API. Uncomment when it becomes stable.
+
+        // ErrorKind::IsADirectory => {
+        //     error!("{:?} is a directory", file);
+        //     exitcode::DATAERR
+        // }
+
         // Permission Error!
         ErrorKind::PermissionDenied => {
-            eprintln!("Can't access {:?}: Permission denied", file,);
+            error!("Can't access {:?}: Permission denied", file);
             exitcode::NOPERM
         }
 
+        // Not UTF-8
+        ErrorKind::InvalidData => {
+            error!(
+                "The file {:?} is not a UTF-8 text file\n\
+                Did you specify a file with a different encoding by accident?",
+                file,
+            );
+            exitcode::DATAERR
+        }
+
         // Other. Just display the name, and exit.
         _ => {
-            eprintln!("{:?} Error while trying to access {:?}", kind, file);
+            error!("{:?} Error while trying to access {:?}", kind, file);
             exitcode::UNAVAILABLE
         }
     }
@@ -134,7 +199,7 @@ fn display_file_error(kind: ErrorKind, cli_options: &Cli, file: &Path) -> i32 {
 fn get_specific_shell<'a>(
     shell: &Shell,
     file_data: &'a ConfigFile,
-) -> Option<&'a EnvironmentVariables> {
+) -> Option<&'a IndexMap<String, EnvVariableValue>> {
     //! Gets the specific environment variables IndexMap for a specific shell.
     //!
     //! ie. This will return the map for `Shell::Zsh`, which looks like this in TOML:
@@ -152,12 +217,23 @@ fn to_shell_source(vars: &EnvironmentVariables, shell: &Shell) -> String {
     //! Converts the hash table of `vars` into a script for the given `shell`.
 
     let mut output = String::new();
-    for (name, raw_value) in vars {
+    for (name, variable_option) in vars {
+        // Check whether the current item is a single environment var or a table of specific shells.
+        let raw_value = match variable_option.clone() {
+            EnvVariableOption::General(v) => v,
+
+            // If it is a shell specific choice, get the correct value for `shell`, and then...
+            EnvVariableOption::Specific(map) => match value_for_specific(shell, map) {
+                Some(v) => v,     // ...extract the `EnvVariableValue` if it exists
+                None => continue, // ...and skip the value if it does not.
+            },
+        };
+
         // Convert an array to a string, but log if it was an array.
-        // Any array are treated
+        // Any arrays are treated as a path.
         let (value, is_path) = match raw_value.clone() {
-            EnvValue::String(s) => (s, false),
-            EnvValue::Array(v) => (v.join(":"), true),
+            EnvVariableValue::String(s) => (s, false),
+            EnvVariableValue::Array(v) => (v.join(":"), true),
         };
 
         // Replace TOML escape codes with the literal representation so that they are correctly used.
@@ -171,6 +247,19 @@ fn to_shell_source(vars: &EnvironmentVariables, shell: &Shell) -> String {
             .replace('\r', r"\r") // Carriage Return
             .replace('\"', "\\\""); // Double Quote
 
+        // Expand tildes
+        let value = shellexpand::tilde(&value);
+
+        // Log each processed variable
+        if log_enabled!(log::Level::Trace) {
+            let variable_log_header = match is_path {
+                true => "PATH EnvVar",
+                false => "EnvVariable",
+            };
+            trace!("{}: {} -> {}", variable_log_header, name, value);
+        };
+
+        // Select the correct form for the chosen shell.
         match shell {
             Shell::Bash | Shell::Zsh => {
                 output += &*format!("export {}=\"{}\";\n", name, value);
@@ -188,6 +277,16 @@ fn to_shell_source(vars: &EnvironmentVariables, shell: &Shell) -> String {
     output
 }
 
+fn value_for_specific(
+    shell: &Shell,
+    map: IndexMap<String, EnvVariableValue>,
+) -> Option<EnvVariableValue> {
+    //! Given a `shell` and a `map` of all specific shell options, get the correct shell `EnvVariableValue`.
+    //! Used by `to_shell_source` to filter the right `EnvVariableOption::Specific` for the current shell.
+    let shell_name = shell.to_possible_value()?.get_name();
+    Some(map.get(shell_name).or_else(|| map.get("_"))?.to_owned())
+}
+
 fn get_file_path_default() -> PathBuf {
     //! Gets the default file path for `xshe.toml` if the `-f`/`--file` option is not set.
 
@@ -200,5 +299,18 @@ fn get_file_path_default() -> PathBuf {
         .into_owned()
         .into();
 
+    info!(
+        "Using default xshe.toml location: {}",
+        xdg_config_home.to_string_lossy()
+    );
+
     xdg_config_home.join("xshe.toml")
 }
+
+fn read_stdin() -> String {
+    //! Read all text from stdin.
+    let mut buffer = String::new();
+    std::io::stdin().lock().read_to_string(&mut buffer).unwrap();
+    debug!("The following input was read from stdin:\n{}", &buffer);
+    buffer
+}
